import os
import sys
import logging
import configparser
import pwd
import attr
import click
import tqdm
import coloredlogs
from . import lib

logger = logging.getLogger(__name__)
current_user = pwd.getpwuid(os.getuid()).pw_name

MB_CONFIG = 'MB_CONFIG'
MB_LOG = 'MB_LOG'
MB_INFO = 'MB_INFO'
MB_DEBUG = 'MB_DEBUG'
MB_TIMINGS = 'MB_TIMINGS'
MB_VERBOSITY = 'MB_VERBOSITY'
MB_QUIET = 'MB_QUIET'
MB_COLORS = 'MB_COLORS'

DEFAULT_CONFIG = '~/musicbot.ini'
DEFAULT_LOG = None
DEFAULT_VERBOSITY = 'warning'
DEFAULT_QUIET = False
DEFAULT_DEBUG = False
DEFAULT_COLORS = True
DEFAULT_TIMINGS = False
DEFAULT_INFO = False


verbosities = {'debug': logging.DEBUG,
               'info': logging.INFO,
               'warning': logging.WARNING,
               'error': logging.ERROR,
               'critical': logging.CRITICAL}

config_option = [click.option('--config', '-c', help='Config file path', type=click.Path(), envvar=MB_CONFIG, default=DEFAULT_CONFIG, show_default=True)]
log_option = [click.option('--log', '-l', help='Log file path', type=click.Path(), envvar=MB_LOG, default=DEFAULT_LOG, show_default=True)]
info_option = [click.option('--info', '-i', help='Same as "--verbosity info"', envvar=MB_INFO, default=DEFAULT_INFO, is_flag=True, show_default=False)]
debug_option = [click.option('--debug', '-d', help='Be very verbose, same as "--verbosity debug" + hide progress bars', envvar=MB_DEBUG, default=DEFAULT_DEBUG, is_flag=True, show_default=True)]
timings_option = [click.option('--timings', '-t', help='Set verbosity to info and show execution timings', envvar=MB_TIMINGS, default=DEFAULT_TIMINGS, is_flag=True, show_default=True)]
verbosity_option = [click.option('--verbosity', '-v', help='Verbosity levels', envvar=MB_VERBOSITY, default=DEFAULT_VERBOSITY, type=click.Choice(verbosities.keys()), show_default=True)]
quiet_option = [click.option('--quiet', '-q', help='Disable progress bars', envvar=MB_QUIET, default=DEFAULT_QUIET, is_flag=True, show_default=True)]

options = config_option + log_option + info_option + debug_option + timings_option + verbosity_option + quiet_option


class TqdmStream:
    @classmethod
    def write(cls, s):
        tqdm.tqdm.write(s, end='', file=sys.stderr)

    @classmethod
    def isatty(cls):
        return True


def check_file_writable(fnm):
    if os.path.exists(fnm):
        # path exists
        if os.path.isfile(fnm):
            return os.access(fnm, os.W_OK)
        return False
    pdir = os.path.dirname(fnm)
    if not pdir:
        pdir = '.'
    return os.access(pdir, os.W_OK)


@attr.s
class Config:
    log = attr.ib(default=DEFAULT_LOG)
    quiet = attr.ib(default=DEFAULT_QUIET)
    debug = attr.ib(default=DEFAULT_DEBUG)
    info = attr.ib(default=DEFAULT_INFO)
    timings = attr.ib(default=DEFAULT_TIMINGS)
    verbosity = attr.ib(default=DEFAULT_VERBOSITY)
    config = attr.ib(default=DEFAULT_CONFIG)
    configfile = attr.ib(default=None, repr=False)
    level = attr.ib(default=verbosities[DEFAULT_VERBOSITY])
    fmt = attr.ib(default="%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s")

    def set(self, config=None, debug=None, info=None, timings=None, quiet=None, verbosity=None, log=None):
        self.config = config if config is not None else os.getenv(MB_CONFIG, DEFAULT_CONFIG)
        self.config = os.path.expanduser(self.config)
        self.configfile = configparser.ConfigParser()
        self.configfile.read(self.config)

        self.log = log if log is not None else os.getenv(MB_LOG, DEFAULT_LOG)
        if self.log:
            self.log = os.path.expanduser(self.log)

        self.quiet = quiet if quiet is not None else lib.str2bool(os.getenv(MB_QUIET, str(DEFAULT_QUIET)))
        self.debug = debug if debug is not None else lib.str2bool(os.getenv(MB_DEBUG, str(DEFAULT_DEBUG)))
        self.info = info if info is not None else lib.str2bool(os.getenv(MB_INFO, str(DEFAULT_INFO)))
        self.timings = timings if timings is not None else lib.str2bool(os.getenv(MB_TIMINGS, str(DEFAULT_TIMINGS)))
        self.verbosity = verbosity if verbosity is not None else os.getenv(MB_VERBOSITY, DEFAULT_VERBOSITY)

        if self.timings or self.info:
            self.verbosity = 'info'
            self.quiet = True
        if self.debug:
            self.verbosity = 'debug'
            self.quiet = True

        self.level = verbosities[self.verbosity]
        coloredlogs.install(level=self.level, fmt=self.fmt, stream=TqdmStream())

        if self.log is not None:
            if check_file_writable(self.log):
                fh = logging.FileHandler(self.log)
                fh.setLevel(logging.DEBUG)
                logging.getLogger().addHandler(fh)
            else:
                logger.warning('No permission to write to %s for current user %s', self.log, current_user)
        logger.debug(self)

    def write(self):
        with open(self.config, 'w') as output_config:
            self.configfile.write(output_config)


config = Config()



import logging
import json
import click
from prettytable import PrettyTable
from musicbot import helpers
from musicbot.music.spotify import spotify_token_option

logger = logging.getLogger(__name__)


@click.group(cls=helpers.GroupWithHelp)
def cli():
    '''Spotify tool'''


def print_tracks(tracks):
    pt = PrettyTable()
    pt.field_names = ["Track", "Artist", "Album"]
    for t in tracks:
        pt.add_row([t['track']['name'], t['track']['artists'][0]['name'], t['track']['album']['name']])
    print(pt)


def print_playlists(playlists):
    pt = PrettyTable()
    pt.field_names = ["Name", "Size"]
    for p in playlists:
        pt.add_row([p['name'], p['tracks']['total']])
    print(pt)


@cli.command()
@helpers.add_options(spotify_token_option)
def playlists(spotify):
    '''List playlists'''
    playlists = spotify.playlists()
    print_playlists(playlists)


@cli.command()
@helpers.add_options(spotify_token_option)
@click.argument("name")
def playlist(name, spotify):
    '''Show playlist'''
    tracks = spotify.playlist(name)
    print_tracks(tracks)


@cli.command()
@helpers.add_options(spotify_token_option + helpers.output_option)
def tracks(spotify, output):
    '''Show tracks'''
    tracks = spotify.tracks()
    if output == 'table':
        print_tracks(tracks)
    elif output == 'json':
        tracks_dict = [{'title': t['track']['name'], 'artist': t['track']['artists'][0]['name'], 'album': t['track']['album']['name']} for t in tracks]
        print(json.dumps(tracks_dict))



import logging
import json
import attr
import click

logger = logging.getLogger(__name__)


rating_choices = [x * 0.5 for x in range(0, 11)]


def sane_rating(ctx, param, value):
    if value is not None and value in rating_choices:
        ctx.params[param.name] = value
        return ctx.params[param.name]
    return None


min_int = 0
max_int = 2147483647

default_name = None
default_filter = None
default_relative = False
default_shuffle = False
default_youtubes = []
default_no_youtubes = []
default_spotifys = []
default_no_spotifys = []
default_formats = []
default_no_formats = []
default_genres = []
default_no_genres = []
default_limit = max_int
default_min_duration = min_int
default_max_duration = max_int
default_min_size = min_int
default_max_size = max_int
default_min_rating = min(rating_choices)
default_max_rating = max(rating_choices)
default_keywords = []
default_no_keywords = []
default_artists = []
default_no_artists = []
default_titles = []
default_no_titles = []
default_albums = []
default_no_albums = []


@attr.s(frozen=True)
class Filter:
    name = attr.ib(default=default_name)
    relative = attr.ib(default=default_relative)
    shuffle = attr.ib(default=default_shuffle)
    youtubes = attr.ib(default=default_youtubes)
    no_youtubes = attr.ib(default=default_no_youtubes)
    spotifys = attr.ib(default=default_spotifys)
    no_spotifys = attr.ib(default=default_no_spotifys)
    formats = attr.ib(default=default_formats)
    no_formats = attr.ib(default=default_no_formats)
    limit = attr.ib(default=default_limit)
    genres = attr.ib(default=default_genres)
    no_genres = attr.ib(default=default_no_genres)
    genres = attr.ib(default=default_genres)
    no_genres = attr.ib(default=default_no_genres)
    min_duration = attr.ib(default=default_min_duration)
    max_duration = attr.ib(default=default_max_duration)
    min_size = attr.ib(default=default_min_size)
    max_size = attr.ib(default=default_max_size)
    min_rating = attr.ib(default=default_min_rating, validator=attr.validators.in_(rating_choices))
    max_rating = attr.ib(default=default_max_rating, validator=attr.validators.in_(rating_choices))
    keywords = attr.ib(default=default_keywords)
    no_keywords = attr.ib(default=default_no_keywords)
    artists = attr.ib(default=default_artists)
    no_artists = attr.ib(default=default_no_artists)
    titles = attr.ib(default=default_titles)
    no_titles = attr.ib(default=default_no_titles)
    albums = attr.ib(default=default_albums)
    no_albums = attr.ib(default=default_no_albums)

    def __attrs_post_init__(self):
        if self.min_size > self.max_size:
            raise ValueError(f"Invalid minimum ({self.min_rating}) or maximum ({self.max_rating}) size")

        if self.min_rating > self.max_rating:
            raise ValueError(f"Invalid minimum ({self.min_rating}) or maximum ({self.max_rating}) rating")

        if self.min_duration > self.max_duration:
            raise ValueError(f"Invalid minimum ({self.min_duration}) or maximum ({self.max_duration}) duration")

        is_bad_formats = set(self.formats).intersection(self.no_formats)
        is_bad_artists = set(self.artists).intersection(self.no_artists)
        is_bad_genres = set(self.genres).intersection(self.no_genres)
        is_bad_albums = set(self.albums).intersection(self.no_albums)
        is_bad_titles = set(self.titles).intersection(self.no_titles)
        is_bad_keywords = set(self.keywords).intersection(self.no_keywords)
        not_empty_set = is_bad_formats or is_bad_artists or is_bad_genres or is_bad_albums or is_bad_titles or is_bad_keywords
        if not_empty_set:
            raise ValueError(f"You can't have duplicates value in filters {self}")
        logger.debug('Filter: %s', self)

    def __repr__(self):
        return self.ordered_dict()

    def diff(self):
        '''Print only differences with default filter'''
        myself = vars(self)
        default = vars(Filter())
        return {k: myself[k] for k in myself if default[k] != myself[k] and k != 'name'}

    def common(self):
        '''Print common values with default filter'''
        myself = vars(self)
        default = vars(Filter())
        return {k: myself[k] for k in myself if default[k] == myself[k] and k != 'name'}

    def ordered_dict(self):
        from collections import OrderedDict
        return OrderedDict([('minDuration', self.min_duration),
                            ('maxDuration', self.max_duration),
                            ('minSize', self.min_size),
                            ('maxSize', self.max_size),
                            ('minRating', self.min_rating),
                            ('maxRating', self.max_rating),
                            ('artists', self.artists),
                            ('noArtists', self.no_artists),
                            ('albums', self.albums),
                            ('noAlbums', self.no_albums),
                            ('titles', self.titles),
                            ('noTitles', self.no_titles),
                            ('genres', self.genres),
                            ('noGenres', self.no_genres),
                            ('formats', self.formats),
                            ('noFormats', self.no_formats),
                            ('keywords', self.keywords),
                            ('noKeywords', self.no_keywords),
                            ('shuffle', self.shuffle),
                            ('relative', self.relative),
                            ('limit', self.limit),
                            ('youtubes', self.youtubes),
                            ('noYoutubes', self.no_youtubes),
                            ('spotifys', self.spotifys),
                            ('noSpotifys', self.no_spotifys)])

    def to_graphql(self):
        return ", ".join([f'{k}: {json.dumps(v)}' for k, v in self.ordered_dict().items()])


options = [
    click.option('--name', help='Filter name', default=default_name),
    click.option('--limit', help='Fetch a maximum limit of music', default=default_limit),
    click.option('--youtubes', help='Select musics with a youtube link', multiple=True, default=default_youtubes),
    click.option('--no-youtubes', help='Select musics without youtube link', multiple=True, default=default_no_youtubes),
    click.option('--spotifys', help='Select musics with a spotifys link', multiple=True, default=default_youtubes),
    click.option('--no-spotifys', help='Select musics without spotifys link', multiple=True, default=default_no_youtubes),
    click.option('--formats', help='Select musics with file format', multiple=True, default=default_formats),
    click.option('--no-formats', help='Filter musics without format', multiple=True, default=default_no_formats),
    click.option('--keywords', help='Select musics with keywords', multiple=True, default=default_keywords),
    click.option('--no-keywords', help='Filter musics without keywords', multiple=True, default=default_no_keywords),
    click.option('--artists', help='Select musics with artists', multiple=True, default=default_artists),
    click.option('--no-artists', help='Filter musics without artists', multiple=True, default=default_no_artists),
    click.option('--albums', help='Select musics with albums', multiple=True, default=default_albums),
    click.option('--no-albums', help='Filter musics without albums', multiple=True, default=default_no_albums),
    click.option('--titles', help='Select musics with titles', multiple=True, default=default_titles),
    click.option('--no-titles', help='Filter musics without titless', multiple=True, default=default_no_titles),
    click.option('--genres', help='Select musics with genres', multiple=True, default=default_genres),
    click.option('--no-genres', help='Filter musics without genres', multiple=True, default=default_no_genres),
    click.option('--min-duration', help='Minimum duration filter (hours:minutes:seconds)', default=default_min_duration),
    click.option('--max-duration', help='Maximum duration filter (hours:minutes:seconds))', default=default_max_duration),
    click.option('--min-size', help='Minimum file size filter (in bytes)', default=default_min_size),
    click.option('--max-size', help='Maximum file size filter (in bytes)', default=default_max_size),
    click.option('--min-rating', help='Minimum rating', default=default_min_rating, show_default=True, callback=sane_rating),
    click.option('--max-rating', help='Maximum rating', default=default_max_rating, show_default=True, callback=sane_rating),
    click.option('--relative', help='Generate relatives paths', default=default_relative, is_flag=True),
    click.option('--shuffle', help='Randomize selection', default=default_shuffle, is_flag=True),
]



import logging
import json
import copy
import os
import click
import taglib


logger = logging.getLogger(__name__)
options = [
    click.option('--keywords', help='Keywords', default=None),
    click.option('--artist', help='Artist', default=None),
    click.option('--album', help='Album', default=None),
    click.option('--title', help='Title', default=None),
    click.option('--genre', help='Genre', default=None),
    click.option('--number', help='Track number', default=None),
    click.option('--rating', help='Rating', default=None),
]
supported_formats = ["mp3", "flac"]


def mysplit(s, delim=','):
    if isinstance(s, list):
        return s
    if s is None:
        return []
    if isinstance(s, str):
        return [x for x in s.split(delim) if x]
    raise ValueError(s)


# pylint: disable-msg=unsupported-membership-test
# pylint: disable-msg=unsubscriptable-object
# pylint: disable-msg=unsupported-assignment-operation
class File:

    def __init__(self, filename, _folder=''):
        self._folder = _folder
        self.handle = taglib.File(filename)
        self.youtube_link = ''
        self.spotify_link = ''

    def __repr__(self):
        return self.path

    def close(self):
        self.handle.close()

    def ordered_dict(self):
        from collections import OrderedDict
        return OrderedDict([('title', self.title),
                            ('album', self.album),
                            ('genre', self.genre),
                            ('artist', self.artist),
                            ('folder', self._folder),
                            ('youtube', self.youtube),
                            ('spotify', self.spotify),
                            ('number', self.number),
                            ('path', self.path),
                            ('rating', self.rating),
                            ('duration', self.duration),
                            ('size', self.size),
                            ('keywords', mysplit(self.keywords, ' '))])

    def __iter__(self):
        yield from self.ordered_dict().items()

    def to_dict(self):
        return dict(self.ordered_dict())

    def to_graphql(self):
        return ", ".join([f'{k}: {json.dumps(v)}' for k, v in self.ordered_dict().items()])

    def to_json(self):
        return json.dumps(self.ordered_dict())

    @property
    def path(self):
        return self.handle.path

    @property
    def folder(self):
        return self._folder

    def __get_first(self, tag, default=''):
        if tag not in self.handle.tags:
            return default
        for item in self.handle.tags[tag]:
            return str(item)
        return default

    def __set_first(self, tag, value, force=False):
        if value is None:
            return
        if tag not in self.handle.tags:
            if force:
                self.handle.tags[tag] = [value]
            return
        del self.handle.tags[tag][0]
        self.handle.tags[tag].insert(0, value)

    @property
    def title(self):
        return self.__get_first('TITLE')

    @title.setter
    def title(self, title):
        self.__set_first('TITLE', title)

    @property
    def album(self):
        return self.__get_first('ALBUM')

    @album.setter
    def album(self, album):
        self.__set_first('ALBUM', album)

    @property
    def artist(self):
        return self.__get_first('ARTIST')

    @artist.setter
    def artist(self, artist):
        self.__set_first('ARTIST', artist)

    @property
    def rating(self):
        s = self.__get_first('FMPS_RATING')
        try:
            n = float(s)
            if n < 0.0:
                return 0.0
            return n * 5.0
        except ValueError:
            return 0.0

    @rating.setter
    def rating(self, rating):
        self.__set_first('FMPS_RATING', rating)

    @property
    def comment(self):
        return self.__get_first('COMMENT')

    @comment.setter
    def comment(self, comment):
        self.__set_first('COMMENT', comment)

    def fix_comment(self, comment):
        self.__set_first('COMMENT', comment, force=True)

    @property
    def description(self):
        return self.__get_first('DESCRIPTION')

    @description.setter
    def description(self, description):
        self.__set_first('DESCRIPTION', description)

    def fix_description(self, description):
        self.__set_first('DESCRIPTION', description, force=True)

    @property
    def genre(self):
        return self.__get_first('GENRE')

    @genre.setter
    def genre(self, genre):
        self.__set_first('GENRE', genre)

    @property
    def number(self):
        s = self.__get_first('TRACKNUMBER')
        try:
            n = int(s)
            if n < 0:
                return -1
            if n > 2 ** 31 - 1:
                logger.warning("%s : invalid number %s", self, n)
                return 0
            return n
        except ValueError:
            return -1

    @number.setter
    def number(self, number):
        self.__set_first('TRACKNUMBER', number)

    @property
    def keywords(self):
        if str(self.handle.path).endswith('.mp3'):
            return self.comment
        if str(self.handle.path).endswith('.flac'):
            if self.comment and not self.description:
                self.fix_description(self.comment)
            return self.description
        return ''

    @keywords.setter
    def keywords(self, keywords):
        if str(self.handle.path).endswith('.mp3'):
            self.comment = keywords
        elif str(self.handle.path).endswith('.flac'):
            self.fix_description(keywords)

    def add_keywords(self, keywords):
        tags = copy.deepcopy(self.keywords)
        for k in keywords:
            if k not in tags:
                tags.append(k)
        if set(self.keywords) != set(tags):
            self.keywords = tags
            self.save()
            return True
        return False

    def delete_keywords(self, keywords):
        tags = copy.deepcopy(self.keywords)
        for k in keywords:
            if k in tags:
                tags.remove(k)
        if set(self.keywords) != set(tags):
            self.keywords = tags
            self.save()
            return True
        return False

    @property
    def duration(self):
        return self.handle.length

    @property
    def size(self):
        return os.path.getsize(self.handle.path)

    @property
    def youtube(self):
        return self.youtube_link

    @property
    def spotify(self):
        return self.spotify_link

    def fingerprint(self, api_key):
        import acoustid
        ids = acoustid.match(api_key, self.path)
        for score, recording_id, title, artist in ids:
            logger.info("score : %s | recording_id : %s | title : %s | artist : %s", score, recording_id, title, artist)
            return recording_id
        return None

    def save(self):
        self.handle.save()



import click
from musicbot.helpers import config_string

DEFAULT_ACOUSTID_API_KEY = None
acoustid_api_key_option = [click.option('--acoustid-api-key', help='AcoustID API Key', default=DEFAULT_ACOUSTID_API_KEY, callback=config_string)]



import click
from logging_tree import printout
from musicbot import helpers
from musicbot.config import config


@click.group(hidden=True, cls=helpers.GroupWithHelp)
def cli():
    '''Config management'''


@cli.command()
def show():
    '''Print default config'''
    print(config)


@cli.command()
def logging():
    '''Show loggers tree'''
    printout()



import time
import os
import logging
import string
import random
import functools
import click
import click_spinner
from click_didyoumean import DYMGroup
from click_aliases import ClickAliasedGroup
from tqdm import tqdm
from .config import config
from .lib import seconds_to_human, find_files, filecount
from .music.file import File, supported_formats

logger = logging.getLogger(__name__)

DEFAULT_MB_CONCURRENCY = 8
concurrency_options = [click.option('--concurrency', help='Number of coroutines', default=DEFAULT_MB_CONCURRENCY, show_default=True)]

DEFAULT_DRY = False
dry_option = [click.option('--dry', help='Take no real action', default=DEFAULT_DRY, is_flag=True, show_default=True)]

DEFAULT_SAVE = False
save_option = [click.option('--save', '-s', help='Save to config file', default=DEFAULT_SAVE, is_flag=True, show_default=True)]

DEFAULT_MB_OUTPUT = 'table'
output_option = [click.option('--output', help='Output format', default=DEFAULT_MB_OUTPUT, show_default=True, type=click.Choice(['table', 'json', 'm3u', 'csv']))]

Red = "\033[0;31;40m"
Green = "\033[0;32;40m"
Yellow = "\033[0;33;40m"
Reset = "\033[0m"


def random_password(size=8):
    alphabet = string.ascii_letters + string.digits
    return ''.join(random.choice(alphabet) for i in range(size))


class GroupWithHelp(DYMGroup, ClickAliasedGroup):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        @click.command('help')
        @click.argument('command', nargs=-1)
        @click.pass_context
        def _help(ctx, command):
            '''Print help'''
            if command:
                argument = command[0]
                c = self.get_command(ctx, argument)
                print(c.get_help(ctx))
            else:
                print(ctx.parent.get_help())
        self.add_command(_help)


def timeit(f):
    @functools.wraps(f)
    def wrapper(*args, **params):
        start = time.time()
        result = f(*args, **params)
        for_human = seconds_to_human(time.time() - start)
        if config.timings:
            logger.info('TIMINGS %s: %s', f.__name__, for_human)
        return result
    return wrapper


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


def config_string(ctx, param, value):
    arg_value = value
    logger.info("%s : try loading with value : %s", param.name, value)

    config_value = config.configfile['DEFAULT'].get(param.name, None)
    logger.info("%s : try loading with config key : %s", param.name, config_value)

    if arg_value:
        value = arg_value

    if config_value:
        if not value or value == param.default:
            value = config_value
        elif arg_value and arg_value != config_value:
            logger.warning("%s : config value %s is not sync with arg value %s", param.name, config_value, arg_value)

    if not value and param.required:
        raise click.BadParameter(f'missing arg or config {param.name} in {config.config}', ctx, param.name, param.name)
    logger.info("%s : final value %s", param.name, value)
    ctx.params[param.name] = value
    return ctx.params[param.name]


def config_list(ctx, param, value):
    arg_value = value
    logger.info("%s : try loading with value : %s", param.name, value)

    config_value = config.configfile['DEFAULT'].get(param.name, None)
    if config_value is not None:
        config_value = tuple(config_value.split(','))
    logger.info("%s : try loading with config key : %s", param.name, config_value)

    if arg_value:
        value = arg_value

    if config_value:
        if not value or value == param.default:
            value = config_value
        elif arg_value and arg_value != config_value:
            logger.warning("%s : config value %s is not sync with arg value %s", param.name, config_value, arg_value)

    if not value and param.required:
        raise click.BadParameter(f'missing arg or config {param.name} in {config.config}', ctx, param.name, param.name)
    logger.info("%s : final value %s", param.name, value)
    ctx.params[param.name] = value
    return ctx.params[param.name]


@timeit
def genfiles(folders):
    with click_spinner.spinner(disable=config.quiet):
        count = 0
        directories = [os.path.abspath(f) for f in folders]
        for d in directories:
            count += filecount(d, supported_formats)
        logger.info("File count: %s", count)
    files = []
    with tqdm(total=count, desc=f"Music listing {folders}", disable=config.quiet) as pbar:
        file_list = find_files(folders, supported_formats)
        music_files = list(file_list)
        for f in music_files:
            try:
                m = File(f[1], f[0])
                files.append(m)
                pbar.update(1)
            except OSError as e:
                logger.error(e)
    return files


folders_argument = [click.argument('folders', nargs=-1, callback=config_list)]



import logging
import json
import click
from prettytable import PrettyTable
from musicbot import helpers, user
from musicbot.config import config

logger = logging.getLogger(__name__)


@click.group(cls=helpers.GroupWithHelp)
def cli():
    '''User management'''


@cli.command('list')
@helpers.add_options(helpers.output_option + user.graphql_admin_option)
def _list(graphql_admin, output):
    '''List users (admin)'''
    a = user.Admin(graphql=graphql_admin)
    users = a.users()
    if output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Email", "Firstname", "Lastname", "Created at", "Updated at"]
        for u in users:
            pt.add_row([u["accountByUserId"]["email"], u["firstName"], u["lastName"], u["createdAt"], u["updatedAt"]])
        print(pt)
    elif output == 'json':
        print(json.dumps(users))


@cli.command(aliases=['new', 'add', 'create'])
@helpers.add_options(user.register_options + helpers.save_option)
def register(save, **kwargs):
    '''Register a new user'''
    u = user.User.register(**kwargs)
    if u.token and save:
        logger.info("saving user infos")
        config.configfile['DEFAULT']['email'] = u.email
        config.configfile['DEFAULT']['password'] = u.password
        config.configfile['DEFAULT']['token'] = u.token
        config.write()


@cli.command(aliases=['delete', 'remove'])
@helpers.add_options(user.auth_options)
def unregister(**kwargs):
    '''Remove a user'''
    u = user.User(**kwargs)
    u.unregister()


@cli.command(aliases=['token'])
@helpers.add_options(user.login_options + helpers.save_option)
def login(save, **kwargs):
    '''Authenticate user'''
    u = user.User(**kwargs)
    print(u.token)
    if u.token and save:
        logger.info("saving user infos")
        config.configfile['DEFAULT']['token'] = u.token
        config.write()



import logging
import click
from musicbot import helpers
from musicbot.music.file import File
from musicbot.music.fingerprint import acoustid_api_key_option

logger = logging.getLogger(__name__)


@click.group(help='''Music file''', cls=helpers.GroupWithHelp)
def cli():
    pass


@cli.command(help='''Print music fingerprint''')
@click.argument('path')
@helpers.add_options(acoustid_api_key_option)
def fingerprint(path, acoustid_api_key):
    f = File(path)
    print(f.fingerprint(acoustid_api_key))



import logging
import json
import click
from slugify import slugify
from musicbot import helpers

logger = logging.getLogger(__name__)


@click.group(help='JSON tools for music libraries', cls=helpers.GroupWithHelp)
def cli():
    pass


@cli.command(help='Diff tracks')
@click.argument('source', type=click.File('r'))
@click.argument('destination', type=click.File('r'))
def diff(source, destination):
    source = json.loads(source.read())
    destination = json.loads(destination.read())
    stopwords = ['the', 'remaster', 'remastered', 'cut', 'part'] + list(map(str, range(1900, 2020)))
    replacements = [['praxis', 'buckethead'], ['lawson-rollins', 'buckethead']]
    source_items = {slugify(f"""{t['artist']}-{t['title']}""", stopwords=stopwords, replacements=replacements) for t in source}
    destination_items = {slugify(f"""{t['artist']}-{t['title']}""", stopwords=stopwords, replacements=replacements) for t in destination}
    differences = source_items.difference(destination_items)
    differences = sorted(differences)
    for difference in differences:
        print(difference)
    print(f"diff : {len(differences)}")



import os
import logging
import click
import acoustid
from humanfriendly import format_timespan
from pytube import YouTube
from pytube.exceptions import VideoUnavailable
from tqdm import tqdm
from musicbot import helpers
from musicbot.config import config
from musicbot.music import youtube
from musicbot.music.file import File
from musicbot.music.youtube import youtube_api_key_option
from musicbot.music.fingerprint import acoustid_api_key_option

logger = logging.getLogger(__name__)


@click.group(cls=helpers.GroupWithHelp)
def cli():
    '''Youtube tool'''


@cli.command()
@helpers.add_options(youtube_api_key_option)
@click.argument('artist')
@click.argument('title')
def search(**kwargs):
    '''Search a youtube link with artist and title'''
    print(youtube.search(**kwargs))


@cli.command()
@click.argument('path')
@helpers.add_options(youtube_api_key_option + acoustid_api_key_option)
def find(path, acoustid_api_key, youtube_api_key):
    '''Search a youtube link with artist and title'''
    f = File(path)
    try:
        file_id = f.fingerprint(acoustid_api_key)
        print(f'Searching for artist {f.artist} and title {f.title} and duration {format_timespan(f.duration)}')
        urls = youtube.search(youtube_api_key, f.artist, f.title, f.duration)
        for url in urls:
            yt_path = None
            try:
                yt = YouTube(url)
                for stream in yt.streams.filter(only_audio=True):
                    with tqdm(total=stream.filesize, desc=f"Testing {url}", disable=config.quiet, leave=False) as pbar:
                        def show_progress_bar(stream, chunk, bytes_remaining):
                            del stream, bytes_remaining
                            pbar.update(len(chunk))  # pylint: disable=cell-var-from-loop
                        yt.register_on_progress_callback(show_progress_bar)
                        yt_path = stream.download()
                        break
                yt_ids = acoustid.match(acoustid_api_key, yt_path)
                yt_id = None
                for _, recording_id, _, _ in yt_ids:
                    yt_id = recording_id
                    break
                if file_id == yt_id:
                    print(f'Found: fingerprint {file_id} | url {url}')
                    break
                if yt_id is not None:
                    print(f'Not exactly found: fingerprint file: {file_id} | yt: {yt_id} | url {url}')
                    break
                print(f'Based only on duration, maybe: {url}')
            except VideoUnavailable:
                pass
            finally:
                if yt_path:
                    os.remove(yt_path)
    except acoustid.WebServiceError as e:
        logger.error(e)


@cli.command()
@click.argument('url')
@helpers.add_options(acoustid_api_key_option)
def fingerprint(url, acoustid_api_key):
    '''Fingerprint a youtube video'''
    path = None
    try:
        yt = YouTube(url)
        stream = yt.streams.filter(only_audio=True).order_by('abr').asc().first()
        with tqdm(total=stream.filesize, desc="Downloading music", disable=config.quiet, leave=False) as pbar:
            def show_progress_bar(stream, chunk, bytes_remaining):
                del stream, bytes_remaining
                pbar.update(len(chunk))
            yt.register_on_progress_callback(show_progress_bar)
            path = stream.download()
        fps = acoustid.match(acoustid_api_key, path)
        for fp in fps:
            print(fp)
            break
    except acoustid.WebServiceError as e:
        logger.error(e)
    finally:
        if path:
            os.remove(path)



import logging
import json
import click
from prettytable import PrettyTable
from musicbot import helpers, lib
from musicbot.config import config

logger = logging.getLogger(__name__)


@click.group(help='Manage folders', cls=helpers.GroupWithHelp)
def cli():
    pass


@cli.command(help='List tracks')
@helpers.add_options(helpers.folders_argument + helpers.output_option)
def tracks(folders, output):
    tracks = helpers.genfiles(folders)
    if output == 'json':
        tracks_dict = [{'title': t.title, 'artist': t.artist, 'album': t.album} for t in tracks]
        print(json.dumps(tracks_dict))
    elif output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Title", "Artist", "Album"]
        for t in tracks:
            pt.add_row([t.title, t.artist, t.album])
        print(pt)
    else:
        raise NotImplementedError


@cli.command(help='Convert all files in folders to mp3')
@helpers.add_options(helpers.folders_argument + helpers.concurrency_options + helpers.dry_option)
def flac2mp3(folders, concurrency, dry):
    import concurrent.futures as cf
    from pydub import AudioSegment
    flac_files = list(lib.find_files(folders, ['flac']))

    pbar = None
    if not config.quiet:
        pbar = click.progressbar(length=len(flac_files), label='Converting musics')

    def convert(flac_path):
        logger.debug('Converting %s', flac_path)
        flac_audio = AudioSegment.from_file(flac_path, "flac")
        mp3_path = flac_path.replace('.flac', '.mp3')
        if not dry:
            flac_audio.export(mp3_path, format="mp3")
        else:
            logger.info("[DRY-RUN] Exporting from %s to %s", flac_path, mp3_path)
        if pbar:
            pbar.update(1)
    # Permit CTRL+C to work as intended
    # import atexit
    # from concurrent.futures.thread import _python_exit
    # atexit.unregister(_python_exit)  # pylint: disable=protected-access
    with cf.ThreadPoolExecutor(max_workers=concurrency) as executor:
        executor.shutdown = lambda wait: None
        futures = [executor.submit(convert, flac_path) for flac_path in flac_files]
        cf.wait(futures)



import logging
import itertools
import spotipy
import click
from musicbot.helpers import config_string

logger = logging.getLogger(__name__)


def sane_spotify(ctx, param, value):
    spotify = config_string(ctx=ctx, param=param, value=value)
    ctx.params['spotify'] = Spotify(spotify)
    return ctx.params['spotify']


spotify_token_option = [click.option('--spotify', help='Spotify token', callback=sane_spotify)]


class Spotify:
    def __init__(self, spotify_token):
        self.sp = spotipy.Spotify(auth=spotify_token)

    def tracks(self):
        offset = 0
        limit = 50
        objects = []
        while True:
            new_objects = self.sp.current_user_saved_tracks(limit=limit, offset=offset)
            objects.append(new_objects['items'])
            logger.info("chunk size: %s", len(new_objects['items']))
            offset += len(new_objects['items'])
            if len(new_objects['items']) < limit:
                break
        return list(itertools.chain(*objects))

    def playlists(self):
        offset = 0
        limit = 50
        objects = []
        while True:
            new_objects = self.sp.current_user_playlists(limit=limit, offset=offset)
            length = len(new_objects['items'])
            objects.append(new_objects['items'])
            logger.info("chunk size: %s", length)
            offset += length
            if length < limit:
                break
        logger.debug("length: %s", offset)
        return list(itertools.chain(*objects))

    def playlist(self, name):
        playlists = self.playlists()
        for p in playlists:
            if p['name'] == name:
                tracks = []
                results = self.sp.playlist(p['id'], fields="tracks,next")
                new_tracks = results['tracks']
                tracks.append(new_tracks['items'])
                while new_tracks['next']:
                    new_tracks = self.sp.next(new_tracks)
                    tracks.append(new_tracks['tracks']['items'])
                return list(itertools.chain(*tracks))
        return []



import logging
import os
import codecs
import csv
import json
import datetime
from shutil import copyfile
from textwrap import indent
import click
import vlc
from prompt_toolkit import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.application import run_in_terminal, get_app
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit import HTML, print_formatted_text
from tqdm import tqdm
from prettytable import PrettyTable
from musicbot import helpers, user, lib
from musicbot.music import mfilter
from musicbot.config import config
from musicbot.music.file import supported_formats


logger = logging.getLogger(__name__)
logging.getLogger("vlc").setLevel(logging.NOTSET)


@click.group(help='''Local music management''', cls=helpers.GroupWithHelp)
@click.pass_context
@helpers.add_options(user.auth_options)
def cli(ctx, **kwargs):
    ctx.obj.u = lambda: user.User.new(**kwargs)


@cli.command(help='''List artists''')
@click.pass_context
@helpers.add_options(helpers.output_option)
def artists(ctx, output):
    if output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Name"]
        for a in ctx.obj.u().artists:
            pt.add_row([a['name']])
        print(pt)
    elif output == 'json':
        print(json.dumps(ctx.obj.u().artists))
    else:
        raise NotImplementedError


@cli.command(help='''Raw query''')
@click.pass_context
@click.argument('query')
def execute(ctx, query):
    print(json.dumps(ctx.obj.u()._post(query)['data']))


@cli.command()
@click.pass_context
def load_filters(ctx):
    '''Load default filters'''
    ctx.obj.u().load_default_filters()


@cli.command(help='''List filters''')
@click.pass_context
@helpers.add_options(helpers.output_option)
def filters(ctx, output):
    if output == 'json':
        print(json.dumps(ctx.obj.u().filters))
    elif output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Name", "Keywords", "No keywords", "Min rating", "Max rating"]
        for f in ctx.obj.u().filters:
            pt.add_row([f['name'], f['keywords'], f['noKeywords'], f['minRating'], f['maxRating']])
        print(pt)
    else:
        raise NotImplementedError


@cli.command('filter', help='''Print a filter''')
@click.pass_context
@helpers.add_options(helpers.output_option)
@click.argument('name')
def _filter(ctx, name, output):
    f = ctx.obj.u().filter(name)
    if output == 'json':
        print(json.dumps(f))
    elif output == 'table':
        print(f)
    else:
        raise NotImplementedError


@cli.command(help='''Generate some stats for music collection with filters''')
@click.pass_context
@helpers.add_options(helpers.output_option + mfilter.options)
def stats(ctx, output, **kwargs):
    mf = mfilter.Filter(**kwargs)
    stats = ctx.obj.u().do_stat(mf)
    if output == 'json':
        print(json.dumps(stats))
    elif output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Stat", "Value"]
        pt.add_row(["Music", stats['musics']])
        pt.add_row(["Artist", stats['artists']])
        pt.add_row(["Album", stats['albums']])
        pt.add_row(["Genre", stats['genres']])
        pt.add_row(["Keywords", stats['keywords']])
        pt.add_row(["Size", lib.bytes_to_human(int(stats['size']))])
        pt.add_row(["Total duration", datetime.timedelta(seconds=int(stats['duration']))])
        print(pt)
    else:
        raise NotImplementedError


@cli.command(help='''List folders''')
@click.pass_context
@helpers.add_options(helpers.output_option)
def folders(ctx, output):
    _folders = ctx.obj.u().folders
    if output == 'json':
        print(json.dumps(_folders))
    elif output == 'table':
        pt = PrettyTable()
        pt.field_names = ["Folder"]
        for f in _folders:
            pt.add_row([f])
        print(pt)
    else:
        raise NotImplementedError


@cli.command(help='''(re)Load musics''')
@helpers.add_options(helpers.folders_argument)
@click.pass_context
def scan(ctx, folders):
    u = ctx.obj.u()
    if not folders:
        folders = u.folders
    files = helpers.genfiles(folders)
    u.bulk_insert(files)


@cli.command(help='''Just list music files''')
@helpers.add_options(helpers.folders_argument)
@click.pass_context
def find(ctx, folders):
    u = ctx.obj.u()
    if not folders:
        folders = u.folders

    files = lib.find_files(folders, supported_formats)
    for f in files:
        print(f[1])


@cli.command(help='''Watch files changes in folders''')
@click.pass_context
def watch(ctx):
    ctx.obj.u().watch()


@cli.command(help='''Clean all musics''')
@click.pass_context
def clean(ctx):
    ctx.obj.u().clean_musics()


@cli.command(help='''Copy selected musics with filters to destination folder''')
@helpers.add_options(helpers.dry_option + mfilter.options)
@click.argument('destination')
@click.pass_context
def sync(ctx, dry, destination, **kwargs):
    logger.info('Destination: %s', destination)
    mf = mfilter.Filter(**kwargs)
    musics = ctx.obj.u().do_filter(mf)

    files = list(lib.all_files(destination))
    logger.info(f"Files : {len(files)}")
    if not files:
        logger.warning("No files found in destination")

    destinations = {f[len(destination) + 1:]: f for f in files}
    logger.info(f"Destinations : {len(destinations)}")
    sources = {m['path'][len(m['folder']) + 1:]: m['path'] for m in musics}
    logger.info(f"Sources : {len(sources)}")
    to_delete = set(destinations.keys()) - set(sources.keys())
    logger.info(f"To delete: {len(to_delete)}")
    if to_delete:
        with tqdm(total=len(to_delete), disable=config.quiet) as pbar:
            for d in to_delete:
                pbar.set_description(f"Deleting musics and playlists: {os.path.basename(destinations[d])}")
                if not dry:
                    try:
                        logger.info("Deleting %s", destinations[d])
                        os.remove(destinations[d])
                    except OSError as e:
                        logger.error(e)
                else:
                    logger.info("[DRY-RUN] False Deleting %s", destinations[d])
                pbar.update(1)

    to_copy = set(sources.keys()) - set(destinations.keys())
    logger.info(f"To copy: {len(to_copy)}")
    if to_copy:
        with tqdm(total=len(to_copy), disable=config.quiet) as pbar:
            for c in sorted(to_copy):
                final_destination = os.path.join(destination, c)
                try:
                    pbar.set_description(f'Copying {os.path.basename(sources[c])} to {destination}')
                    if not dry:
                        logger.info("Copying %s to %s", sources[c], final_destination)
                        os.makedirs(os.path.dirname(final_destination), exist_ok=True)
                        copyfile(sources[c], final_destination)
                    else:
                        logger.info("[DRY-RUN] False Copying %s to %s", sources[c], final_destination)
                    pbar.update(1)
                except KeyboardInterrupt:
                    logger.debug(f"Cleanup {final_destination}")
                    os.remove(final_destination)
                    raise

    import shutil
    for d in lib.empty_dirs(destination):
        if any(e in d for e in lib.exceptions):
            logger.debug(f"Invalid path {d}")
            continue
        if not dry:
            shutil.rmtree(d)
        logger.info("[DRY-RUN] Removing empty dir %s", d)


@cli.command(help='''Check music files consistency''')
@click.pass_context
@helpers.add_options(helpers.folders_argument)
def consistency(ctx, folders):
    if not folders:
        folders = ctx.obj.u().folders

    musics = helpers.genfiles(folders)
    pt = PrettyTable()
    pt.field_names = ["Inconsistency", "File"]
    for m in musics:
        try:
            if m.path.endswith('.flac'):
                if m.comment and not m.description:
                    pt.add_row(['Comment (' + m.comment + ') used in flac', m.path])
            if m.path.endswith('.mp3'):
                if m.description and not m.comment:
                    pt.add_row(['Description (' + m.description + ') used in mp3 : ', m.path])
            if not m.title:
                pt.add_row(["No title  : '" + m.title + "' on ", m.path])
            filename = os.path.basename(m.path)
            if filename == f"{str(m.number).zfill(2)} - {m.title}.mp3":
                continue
            if filename == f"{str(m.number).zfill(2)} - {m.title}.flac":
                continue
            pt.add_row([f"Invalid title format, '{filename}' should start by '{str(m.number).zfill(2)} - {m.title}'", m.path])
            if m.artist not in m.path:
                pt.add_row(["Artist invalid : " + m.artist + " is not in ", m.path])
            if m.genre == '':
                pt.add_row(["No genre : ", m.path])
            if m.album == '':
                pt.add_row(["No album : ", m.path])
            if m.artist == '':
                pt.add_row(["No artist : ", m.path])
            if m.rating == 0.0:
                pt.add_row(["No rating : ", m.path])
            if m.number == -1:
                pt.add_row(["Invalid track number : ", m.path])
        except OSError:
            pt.add_row(["Could not open file : ", m.path])
    print(pt)


@cli.command(help='Music player')
@helpers.add_options(user.auth_options)
@helpers.add_options(mfilter.options)
@click.pass_context
def play(ctx, email, password, token, graphql, **kwargs):
    try:
        ctx.obj.u = lambda: user.User.new(email=email, password=password, token=token, graphql=graphql)
        mf = mfilter.Filter(**kwargs)
        p = ctx.obj.u().do_filter(mf)
        if not p:
            logger.warning('Empty playlist')
            return
        instance = vlc.Instance()
        songs = [song['path'] for song in p]
        player = instance.media_list_player_new()
        media_list = instance.media_list_new(songs)
        player.set_media_list(media_list)
        bindings = KeyBindings()

        @bindings.add('p')
        def _play_binding():
            def play():
                """Play song"""
                player.play()
            run_in_terminal(play)

        @bindings.add('q')
        def _quit_binding(event):
            player.pause()
            event.app.exit()

        @bindings.add('s')
        def _pause_binding():
            player.pause()

        @bindings.add('l')
        def _playlist_binding():
            def playlist():
                """List songs"""
                media_player = player.get_media_player()
                media = media_player.get_media()
                media.parse()
                artist = media.get_meta(vlc.Meta.Artist)
                album = media.get_meta(vlc.Meta.Album)
                title = media.get_meta(vlc.Meta.Title)
                for s in p:
                    if s['artist'] == artist and s['title'] == title and s['album'] == album:
                        print_formatted_text(f'''> {s['path']}''')
                    else:
                        print_formatted_text(s['path'])
                print_formatted_text('------------------------------------------')
            run_in_terminal(playlist)

        @bindings.add('right')
        def _next_binding():
            player.next()

        @bindings.add('left')
        def _previous_binding():
            player.previous()

        def bottom_toolbar():
            media_player = player.get_media_player()
            media = media_player.get_media()
            media.parse()
            media_time = lib.seconds_to_human(round(media_player.get_time() / 1000))
            media_length = lib.seconds_to_human(round(media_player.get_length() / 1000))
            artist = media.get_meta(vlc.Meta.Artist)
            album = media.get_meta(vlc.Meta.Album)
            title = media.get_meta(vlc.Meta.Title)
            current = f'({media_time} / {media_length}) {artist} - {album} - {title}'
            get_app().invalidate()
            return HTML(f'Current song: {current}')

        player.play()
        print_formatted_text(HTML('Bindings: q = quit | p = play | s = pause/continue | right = next song | left = previous song | l = playlist'))

        root_container = HSplit(
            [Window(FormattedTextControl(lambda: bottom_toolbar, style='class:bottom-toolbar.text'), style='class:bottom-toolbar')]
        )
        layout = Layout(root_container)
        app = Application(layout=layout, key_bindings=bindings)
        app.run()
    except NameError:
        logger.critical("Your VLC version may be outdated: %s", vlc.libvlc_get_version())


@cli.command(help='''Generate a new playlist''')
@click.pass_context
@helpers.add_options(helpers.dry_option + mfilter.options + helpers.output_option)
@click.argument('path', type=click.File('w'), default='-')
def playlist(ctx, output, path, dry, **kwargs):
    mf = mfilter.Filter(**kwargs)
    if output == 'm3u':
        p = ctx.obj.u().playlist(mf)
        if not dry:
            print(p, file=path)
        else:
            logger.info('DRY RUN: Writing playlist to %s with content:\n%s', path, p)
    else:
        tracks = ctx.obj.u().do_filter(mf)
        if output == 'json':
            print(json.dumps(tracks), file=path)
        elif output == 'table':
            pt = PrettyTable()
            pt.field_names = [
                "Title",
                "Album",
                "Artist",
            ]
            for t in tracks:
                pt.add_row([t['title'], t['album'], t['artist']])
            print(pt, file=path)
        elif output == 'csv':
            folders = ctx.obj.u().folders
            logger.info('Scanning folders: %s', folders)
            files = helpers.genfiles(folders)

            musicwriter = csv.writer(path, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
            for f in files:
                musicwriter.writerow(f.ordered_dict().values())


@cli.command(help='''Generate bests playlists with some rules''')
@click.pass_context
@helpers.add_options(helpers.dry_option + mfilter.options)
@click.argument('path', type=click.Path(exists=True))
@click.option('--prefix', envvar='MB_PREFIX', help="Append prefix before each path (implies relative)", default='')
@click.option('--suffix', envvar='MB_SUFFIX', help="Append this suffix to playlist name", default='')
def bests(ctx, dry, path, prefix, suffix, **kwargs):
    if prefix:
        kwargs['relative'] = True
    mf = mfilter.Filter(**kwargs)
    playlists = ctx.obj.u().bests(mf)
    with tqdm(total=len(playlists), disable=config.quiet) as pbar:
        for p in playlists:
            playlist_filepath = os.path.join(path, p['name'] + suffix + '.m3u')
            pbar.set_description(f"Best playlist ${prefix} ${suffix}: {os.path.basename(playlist_filepath)}")
            content = indent(p['content'], prefix, lambda line: line != '#EXTM3U')
            if not dry:
                try:
                    with codecs.open(playlist_filepath, 'w', "utf-8-sig") as playlist_file:
                        logger.debug('Writing playlist to %s with content:\n%s', playlist_filepath, content)
                        playlist_file.write(content)
                except (FileNotFoundError, LookupError, ValueError, UnicodeError) as e:
                    logger.warning(f'Unable to write playlist to {playlist_filepath} because of {e}')
            else:
                logger.info('DRY RUN: Writing playlist to %s with content:\n%s', playlist_filepath, content)
            pbar.update(1)



import os
import re
import sys
import logging
import humanfriendly

logger = logging.getLogger(__name__)

output_types = ["list", "json"]
exceptions = ['.Spotlight-V100', '.zfs', 'Android', 'LOST.DIR']
default_output_type = 'json'


def str2bool(val):
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return 1
    if val in ('n', 'no', 'f', 'false', 'off', '0'):
        return 0
    raise ValueError("invalid truth value %r" % (val,))


def bytes_to_human(b):
    return humanfriendly.format_size(b)


def seconds_to_human(s):
    import datetime
    return str(datetime.timedelta(seconds=s))


def empty_dirs(root_dir, recursive=True):
    dirs_list = []
    for root, dirs, files in os.walk(root_dir, topdown=False):
        if recursive:
            all_subs_empty = True
            for sub in dirs:
                full_sub = os.path.join(root, sub)
                if full_sub not in dirs_list:
                    all_subs_empty = False
                    break
        else:
            all_subs_empty = (not dirs)
        if all_subs_empty and is_empty(files):
            dirs_list.append(root)
            yield root


def is_empty(files):
    return len(files) == 0


def raise_limits():
    import resource
    try:
        _, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        logger.info("Current limits, soft and hard : %s %s", _, hard)
        resource.setrlimit(resource.RLIMIT_NOFILE, (hard, hard))
        return True
    except (ValueError, OSError) as e:
        logger.critical('You may need to check ulimit parameter: %s', e)
        return False


def restart():
    python = sys.executable
    print(f'Restarting myself: {python} {sys.argv}')
    # only works on linux, not windows with WSL
    # os.execl(python, python, * sys.argv)
    # permit to exit from a thread
    os._exit(0)


def find_files(directories, supported_formats):
    directories = [os.path.abspath(d) for d in directories]
    for directory in directories:
        for root, _, files in os.walk(directory):
            if any(e in root for e in exceptions):
                logger.debug(f"Invalid path {root}")
                continue
            for basename in files:
                filename = os.path.join(root, basename)
                if filename.endswith(tuple(supported_formats)):
                    yield (directory, filename)


def scantree(path, supported_formats):
    try:
        if '/.' in path:
            return
        for entry in os.scandir(path):
            if entry.is_dir(follow_symlinks=False):
                yield from scantree(entry.path, supported_formats)
            else:
                if entry.name.endswith(tuple(supported_formats)):
                    yield entry
    except PermissionError as e:
        logger.error(e)


def filecount(path, supported_formats):
    return len(list(scantree(path, supported_formats)))


def all_files(directory):
    for root, _, files in os.walk(directory):
        if any(e in root for e in exceptions):
            logger.debug(f"Invalid path {root}")
            continue
        for basename in files:
            yield os.path.join(root, basename)


def first(iterable, default=None):
    if iterable:
        if isinstance(iterable, str):
            return iterable
        for item in iterable:
            return item
    return default


def num(s):
    try:
        return int(s)
    except ValueError:
        return float(s)


def duration_to_seconds(duration):
    if re.match(r'\d+s', duration):
        return int(duration[:-1])
    if re.match(r'\d+m', duration):
        return int(duration[:-1]) * 60
    if re.match(r'\d+h', duration):
        return int(duration[:-1]) * 3600
    raise ValueError(duration)


def seconds_to_str(duration):
    import datetime
    return str(datetime.timedelta(seconds=duration))


default_checks = ['keywords', 'strict_title', 'title', 'path',
                  'genre', 'album', 'artist', 'rating', 'number']



import base64
import json
import logging
import functools
import requests
import click
import click_spinner
from tqdm import tqdm
from musicbot import helpers
from musicbot.config import config
from musicbot.helpers import config_string
from musicbot.music import file, mfilter

logger = logging.getLogger(__name__)
DEFAULT_TOKEN = None
token_option = [click.option('--token', '-t', help='User token', default=DEFAULT_TOKEN, callback=config_string)]

DEFAULT_EMAIL = None
email_option = [click.option('--email', '-e', help='User email', default=DEFAULT_EMAIL, callback=config_string)]

DEFAULT_PASSWORD = None
password_option = [click.option('--password', '-p', help='User password', default=DEFAULT_PASSWORD, callback=config_string)]

DEFAULT_FIRST_NAME = None
first_name_option = [click.option('--first-name', help='User first name', default=DEFAULT_FIRST_NAME, callback=config_string, show_default=True)]

DEFAULT_LAST_NAME = None
last_name_option = [click.option('--last-name', help='User last name', default=DEFAULT_FIRST_NAME, callback=config_string, show_default=True)]

DEFAULT_GRAPHQL_ADMIN = 'http://127.0.0.1:5001/graphql'
graphql_admin_option = [click.option('--graphql-admin', help='GraphQL endpoint', default=DEFAULT_GRAPHQL_ADMIN, callback=config_string, show_default=True)]

DEFAULT_GRAPHQL = 'http://127.0.0.1:5000/graphql'
graphql_option = [click.option('--graphql', help='GraphQL endpoint', default=DEFAULT_GRAPHQL, callback=config_string, show_default=True)]

register_options = email_option + password_option + first_name_option + last_name_option + graphql_option
login_options = email_option + password_option + graphql_option
auth_options = login_options + token_option


class MusicbotError(Exception):
    pass


class FailedAuthentication(MusicbotError):
    pass


class FailedRequest(MusicbotError):
    pass


class GraphQL:
    def __init__(self, graphql, headers=None):
        self.graphql = graphql
        self.headers = headers

    def _post(self, query, failure=None):
        logger.debug(query)
        response = requests.post(self.graphql, json={'query': query}, headers=self.headers)
        logger.debug(response)
        json_response = response.json()
        logger.debug(json_response)
        if response.status_code != 200:
            failure = failure if failure is not None else FailedRequest(f"Query failed: {json_response}")
            raise failure
        if 'errors' in json_response and json_response['errors']:
            errors = [e['message'] for e in json_response['errors']]
            failure = failure if failure is not None else FailedRequest(f"Query failed: {errors}")
            raise failure
        return json_response


class Admin(GraphQL):
    @helpers.timeit
    def __init__(self, graphql=None):
        graphql = graphql if graphql is not None else DEFAULT_GRAPHQL_ADMIN
        GraphQL.__init__(self, graphql=graphql)

    @helpers.timeit
    def users(self):
        query = """
        {
          usersList{
              firstName,
              lastName,
              createdAt,
              updatedAt,
              accountByUserId{
                email
              }
          }
        }"""
        return self._post(query)['data']['usersList']


class User(GraphQL):
    @helpers.timeit
    def __init__(self, graphql=None, email=None, password=None, token=None):
        self.graphql = graphql if graphql is not None else DEFAULT_GRAPHQL
        self.email = email
        self.password = password
        self.token = token
        self.authenticated = False

        if self.token:
            logger.debug("using token : %s", self.token)
        elif self.email and self.password:
            query = f"""
            mutation
            {{
                authenticate(input: {{email: "{self.email}", password: "{self.password}"}})
                {{
                    jwtToken
                }}
            }}"""
            self.headers = None

            response = self._post(query, failure=FailedAuthentication(f"Authentication failed for email {self.email}"))
            try:
                self.token = response['data']['authenticate']['jwtToken']
            except KeyError:
                raise FailedAuthentication(f"Invalid response received : {response}")
            if not self.token:
                raise FailedAuthentication(f"Invalid token received : {self.token}")
        else:
            raise FailedAuthentication("No credentials or token provided")
        self.authenticated = True
        GraphQL.__init__(self, graphql=graphql, headers={"Authorization": f"Bearer {self.token}"})

    @classmethod
    @functools.lru_cache(maxsize=None)
    @helpers.timeit
    def new(cls, **kwargs):
        self = User(**kwargs)
        return self

    @helpers.timeit
    def load_default_filters(self):
        query = """
        mutation
        {
            default:           createFilter(input: {filter: {name: "default"                                                                           }}){clientMutationId}
            no_artist_set:     createFilter(input: {filter: {name: "no artist set",     artists:    ""                                                 }}){clientMutationId}
            no_album_set:      createFilter(input: {filter: {name: "no album set",      albums:     ""                                                 }}){clientMutationId}
            no_title_set:      createFilter(input: {filter: {name: "no title set",      titles:     ""                                                 }}){clientMutationId}
            no_genre_set:      createFilter(input: {filter: {name: "no genre set",      genres:     ""                                                 }}){clientMutationId}
            youtube_not_found: createFilter(input: {filter: {name: "youtube not found", youtubes:   ["not found"]                                      }}){clientMutationId}
            spotify_not_found: createFilter(input: {filter: {name: "spotify not found", spotifys:   ["not found"]                                      }}){clientMutationId}
            no_youtube_links:  createFilter(input: {filter: {name: "no youtube links",  youtubes:   []                                                 }}){clientMutationId}
            no_spotify_links:  createFilter(input: {filter: {name: "no spotify links",  spotifys:   ["not found"]                                      }}){clientMutationId}
            no_rating:         createFilter(input: {filter: {name: "no rating",         minRating:  0.0, maxRating: 0.0                                }}){clientMutationId}
            bests_40:          createFilter(input: {filter: {name: "best (4.0+)",       minRating:  4.0, noKeywords: ["cutoff", "bad", "demo", "intro"]}}){clientMutationId}
            bests_45:          createFilter(input: {filter: {name: "best (4.5+)",       minRating:  4.5, noKeywords: ["cutoff", "bad", "demo", "intro"]}}){clientMutationId}
            bests_50:          createFilter(input: {filter: {name: "best (5.0+)",       minRating:  5.0, noKeywords: ["cutoff", "bad", "demo", "intro"]}}){clientMutationId}
            no_live:           createFilter(input: {filter: {name: "no live",           noKeywords: ["live"]                                           }}){clientMutationId}
            only_live:         createFilter(input: {filter: {name: "only live",         keywords:   ["live"]                                           }}){clientMutationId}
        }"""
        return self._post(query)

    @helpers.timeit
    def playlist(self, mf=None):
        mf = mf if mf is not None else mfilter.Filter()
        query = f"""
        {{
            playlist({mf.to_graphql()})
        }}"""
        return self._post(query)['data']['playlist']

    @helpers.timeit
    def bests(self, mf=None):
        mf = mf if mf is not None else mfilter.Filter()
        query = f"""
        {{
            bests({mf.to_graphql()})
            {{
                nodes
                {{
                    name,
                    content
                }}
            }}
        }}"""
        return self._post(query)['data']['bests']['nodes']

    @helpers.timeit
    def do_filter(self, mf=None):
        mf = mf if mf is not None else mfilter.Filter()
        if mf.name:
            kwargs = self.filter(mf.name)
            print(kwargs)
            mf = mfilter.Filter(**kwargs)

        query = f"""
        {{
            doFilter({mf.to_graphql()})
            {{
                nodes
                {{
                    title,
                    album,
                    genre,
                    artist,
                    folder,
                    youtube,
                    spotify,
                    number,
                    path,
                    rating,
                    duration,
                    size,
                    keywords
                }}
            }}
        }}"""
        return self._post(query)['data']['doFilter']['nodes']

    @helpers.timeit
    def do_stat(self, mf=None):
        mf = mf if mf is not None else mfilter.Filter()
        query = f"""
        {{
            doStat({mf.to_graphql()})
            {{
              musics,
              artists,
              albums,
              genres,
              keywords,
              size,
              duration
            }}
        }}"""
        return self._post(query)['data']['doStat']

    @helpers.timeit
    def upsert_music(self, music):
        query = f"""
        mutation
        {{
            upsertMusic(input: {{{music.to_graphql()}}})
            {{
                clientMutationId
            }}
        }}"""
        return self._post(query)

    @helpers.timeit
    def bulk_insert(self, musics):
        if not musics:
            logger.info("no musics to insert")
            return None
        if config.debug:
            with tqdm(total=len(musics), desc="inserting music one by one") as pbar:
                for music in musics:
                    logger.debug("inserting %s", music)
                    self.upsert_music(music)
                    pbar.update(1)
            return None

        j = json.dumps([m.to_dict() for m in musics])
        b64 = j.encode('utf-8')
        data = base64.b64encode(b64)
        query = f'''
        mutation
        {{
            bulkInsert(input: {{data: "{data.decode()}"}})
            {{
                clientMutationId
            }}
        }}'''
        with click_spinner.spinner(disable=config.quiet):
            return self._post(query)

    @property
    @functools.lru_cache(maxsize=None)
    @helpers.timeit
    def folders(self):
        query = """
        {
            foldersList
        }"""
        return self._post(query)['data']['foldersList']

    @property
    @functools.lru_cache(maxsize=None)
    @helpers.timeit
    def artists(self):
        query = """
        {
          artistsTreeList {
            name
            albums {
              name
              musics {
                folder
                name
                path
              }
            }
          }
        }"""
        return self._post(query)['data']['artistsTreeList']

    @property
    @functools.lru_cache(maxsize=None)
    @helpers.timeit
    def genres(self):
        query = """
        {
          genresTreeList {
            name
          }
        }"""
        return self._post(query)['data']['genresTreeList']

    @functools.lru_cache(maxsize=None)
    @helpers.timeit
    def filter(self, name):
        default_filter = mfilter.Filter()
        filter_members = ','.join(default_filter.ordered_dict().keys())
        query = f"""
        {{
            filtersList(filter: {{name: {{equalTo: "{name}"}}}})
            {{
                name,
                {filter_members}
            }}
        }}"""
        return self._post(query)['data']['filtersList'][0]

    @property
    @functools.lru_cache(maxsize=None)
    @helpers.timeit
    def filters(self):
        default_filter = mfilter.Filter()
        filter_members = ','.join(default_filter.ordered_dict().keys())
        query = f"""
        {{
            filtersList
            {{
                name,
                {filter_members}
            }}
        }}"""
        return self._post(query)['data']['filtersList']

    def watch(self):
        from watchdog.observers import Observer
        from watchdog.events import PatternMatchingEventHandler
        import time

        class MusicWatcherHandler(PatternMatchingEventHandler):
            patterns = []

            def __init__(self, user):
                super().__init__()
                self.user = user
                MusicWatcherHandler.patterns = ['*.' + f for f in file.supported_formats]

            def on_modified(self, event):
                self.update_music(event.src_path)

            def on_created(self, event):
                self.update_music(event.src_path)

            def on_deleted(self, event):
                logger.debug('Deleting entry in DB for: %s %s', event.src_path, event.event_type)
                self.user.delete_music(event.src_path)

            def on_moved(self, event):
                logger.debug('Moving entry in DB for: %s %s', event.src_path, event.event_type)
                self.user.delete_music(event.src_path)
                self.update_music(event.dest_path)

            def update_music(self, path):
                for folder in self.user.folders:
                    if path.startswith(folder) and path.endswith(tuple(file.supported_formats)):
                        logger.debug('Creating/modifying DB for: %s', path)
                        f = file.File(path, folder)
                        self.user.upsert_music(f)
                        return

        logger.info('Watching: %s', self.folders)
        event_handler = MusicWatcherHandler(self)
        observer = Observer()
        for f in self.folders:
            observer.schedule(event_handler, f, recursive=True)
        observer.start()
        try:
            while True:
                time.sleep(50)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()

    @classmethod
    @helpers.timeit
    def register(cls, graphql=None, first_name=None, last_name=None, email=None, password=None):
        first_name = first_name if first_name is not None else DEFAULT_FIRST_NAME
        last_name = last_name if last_name is not None else DEFAULT_LAST_NAME
        email = email if email is not None else DEFAULT_EMAIL
        password = password if password is not None else DEFAULT_PASSWORD

        if email is None:
            raise click.BadParameter('Missing value for email')
        if password is None:
            raise click.BadParameter('Missing value for password')

        query = f"""
        mutation
        {{
            registerUser(input: {{firstName: "{first_name}", lastName: "{last_name}", email: "{email}", password: "{password}"}})
            {{
                clientMutationId
            }}
        }}"""
        logger.debug(query)
        response = requests.post(graphql, json={'query': query})
        json_response = response.json()
        logger.debug(json_response)
        if response.status_code != 200:
            raise FailedAuthentication(f"Cannot create user: {email}")
        if 'errors' in json_response and json_response['errors']:
            errors = [e['message'] for e in json_response['errors']]
            raise FailedAuthentication(f"Cannot create user: {errors}")
        return User(graphql, email, password)

    @helpers.timeit
    def unregister(self):
        query = """
        mutation
        {
            unregisterUser(input: {})
            {
                clientMutationId
            }
        }"""
        result = self._post(query, failure=FailedAuthentication(f"Cannot delete user {self.email}"))
        self.authenticated = False
        return result

    @helpers.timeit
    def delete_music(self, path):
        query = f"""
        mutation
        {{
            deleteMusic(input: {{path: "{path}"}})
            {{
                clientMutationId
            }}
        }}"""
        return self._post(query)

    @helpers.timeit
    def clean_musics(self):
        query = """
        mutation
        {
            deleteAllMusic(input: {})
            {
                clientMutationId
            }
        }"""
        return self._post(query)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['musicbot', 'musicbot.commands', 'musicbot.music']

package_data = \
{'': ['*'], 'musicbot': ['schema/*']}

install_requires = \
['attrdict>=2.0,<3.0',
 'attrs>=19.1,<20.0',
 'click-didyoumean>=0.0.3,<0.0.4',
 'click-man>=0.4.1,<0.5.0',
 'click-spinner>=0.1.8,<0.2.0',
 'click>=7.0,<8.0',
 'click_aliases>=1.0.1,<2.0.0',
 'coloredlogs>=14.0,<15.0',
 'feedparser>=5.2.1,<6.0.0',
 'flask>=1.1.1,<2.0.0',
 'graphql-py>=0.7.1,<0.8.0',
 'humanfriendly>=8.1,<9.0',
 'isodate>=0.6.0,<0.7.0',
 'isort[pyproject]>=4.3,<5.0',
 'logging-tree>=1.8,<2.0',
 'mps-youtube>=0.2.8,<0.3.0',
 'prettytable>=0.7.2,<0.8.0',
 'prompt_toolkit>=3.0.2,<4.0.0',
 'psycopg2-binary>=2.8,<3.0',
 'pyacoustid>=1.1,<2.0',
 'pydub>=0.23.1,<0.24.0',
 'pytaglib>=1.4,<2.0',
 'python-dotenv>=0.12.0,<0.13.0',
 'python-slugify>=4.0.0,<5.0.0',
 'python-vlc>=3.0,<4.0',
 'pytube3>=9.6.4,<10.0.0',
 'spotipy>=2.10.0,<3.0.0',
 'tqdm>=4.31,<5.0',
 'watchdog>=0.10.0,<0.11.0',
 'youtube-dl>=2020.1.24,<2021.0.0']

entry_points = \
{'console_scripts': ['musicbot = musicbot.cli:main']}

setup_kwargs = {
    'name': 'musicbot',
    'version': '0.2.0',
    'description': 'Music swiss army knife',
    'long_description': '========\nMusicBot\n========\n+---------------+-----------------+\n|     Tools     |      Result     |\n+===============+=================+\n|     Codacy    |    |codacy|     |\n+---------------+-----------------+\n|     Coverage  |   |coverage|    |\n+---------------+-----------------+\n\n.. |codacy| image:: https://api.codacy.com/project/badge/Grade/621acf3309b24c538c40824f9af467de\n   :target: https://www.codacy.com/app/AdrienPensart/musicbot?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=AdrienPensart/musicbot&amp;utm_campaign=Badge_Grade\n   :alt: Codacy badge\n.. |coverage| image:: https://github.com/AdrienPensart/musicbot/blob/master/doc/coverage.svg\n   :alt: Coverage badge\n\nDescription\n-----------\nCLI / API / Website to manipulate music and create smart playlists, and play it !\n\nIt uses poetry and pyenv tools to manage project life and docker to test it.\n\nUnder the hood, there is a postgraphile API frontend backed by a postgresql database, it allow us\nto have users and security integrated.\n\nDev Environment\n------------\n\n.. code-block:: bash\n\n  sudo apt install -y vlc libtag1-dev postgresql-server-dev-all ffmpeg python3-pip docker.io libchromaprint-tools libbz2-dev libsqlite3-dev llvm libncurses5-dev libncursesw5-dev tk-dev liblzma-dev libssl-dev libreadline-dev\n  sudo usermod -aG docker $(whoami)\n\n  git clone https://github.com/AdrienPensart/musicbot.git\n  cd musicbot\n\n  https://pyenv.run | bash\n  pyenv install --verbose $(cat .python-version) -ks\n  pyenv global $(cat .python-version)\n\n  # you should put that at your shell startup\n  eval "$(pyenv init -)"\n\n  python <(curl -sSL https://raw.githubusercontent.com/sdispater/poetry/master/get-poetry.py) --preview\n  poetry install\n\nTesting\n------------\n\n.. code-block:: bash\n\n  poetry run pytest\n  poetry run coverage-badge -f -o doc/coverage.svg\n\nHow to use\n------------\n\n.. code-block:: bash\n\n  poetry run docker-compose build --parallel\n  poetry run docker-compose up\n  musicbot user create --email your_email --password your_password --save\n  musicbot local scan /tests/fixtures/folder1 /tests/fixtures/folder2\n  musicbot local tracks\n\nLinting\n------------\n\n.. code-block:: bash\n\n  poetry run pylint musicbot tests\n\nDocumentation\n------------\n\n.. code-block:: bash\n\n  poetry run doc/gen.sh\n\nCommands\n--------\n.. code-block::\n\n  Usage: musicbot [OPTIONS] COMMAND [ARGS]...\n  \n    Music swiss knife, new gen.\n  \n  Options:\n    -V, --version                                       Show the version and exit.\n    -c, --config PATH                                   Config file path  [default: ~/musicbot.ini]\n    -l, --log PATH                                      Log file path\n    -i, --info                                          Same as "--verbosity info"\n    -d, --debug                                         Be very verbose, same as "--verbosity debug" + hide progress bars  [default: False]\n    -t, --timings                                       Set verbosity to info and show execution timings  [default: False]\n    -v, --verbosity [debug|info|warning|error|critical]\n                                                        Verbosity levels  [default: warning]\n    -q, --quiet                                         Disable progress bars  [default: False]\n    -h, --help                                          Show this message and exit.\n  \n  Commands:\n    folder   Manage folders\n    help     Print help\n    json     JSON tools for music libraries\n    local    Local music management\n    music    Music file\n    spotify  Spotify tool\n    user     User management\n    version  Print version\n    youtube  Youtube tool\n\n\nmusicbot folder\n***************\n.. code-block::\n\n  Usage: musicbot folder [OPTIONS] COMMAND [ARGS]...\n  \n    Manage folders\n  \n  Options:\n    -h, --help  Show this message and exit.\n  \n  Commands:\n    flac2mp3  Convert all files in folders to mp3\n    help      Print help\n    tracks    List tracks\n\n\nmusicbot folder flac2mp3\n************************\n.. code-block::\n\n  Usage: musicbot folder flac2mp3 [OPTIONS] [FOLDERS]...\n  \n    Convert all files in folders to mp3\n  \n  Options:\n    --concurrency INTEGER  Number of coroutines  [default: 8]\n    --dry                  Take no real action  [default: False]\n    -h, --help             Show this message and exit.\n\n\nmusicbot folder tracks\n**********************\n.. code-block::\n\n  Usage: musicbot folder tracks [OPTIONS] [FOLDERS]...\n  \n    List tracks\n  \n  Options:\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    -h, --help                     Show this message and exit.\n\n\nmusicbot help\n*************\n.. code-block::\n\n  Usage: musicbot help [OPTIONS] [COMMAND]...\n  \n    Print help\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot json\n*************\n.. code-block::\n\n  Usage: musicbot json [OPTIONS] COMMAND [ARGS]...\n  \n    JSON tools for music libraries\n  \n  Options:\n    -h, --help  Show this message and exit.\n  \n  Commands:\n    diff  Diff tracks\n    help  Print help\n\n\nmusicbot json diff\n******************\n.. code-block::\n\n  Usage: musicbot json diff [OPTIONS] SOURCE DESTINATION\n  \n    Diff tracks\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot local\n**************\n.. code-block::\n\n  Usage: musicbot local [OPTIONS] COMMAND [ARGS]...\n  \n    Local music management\n  \n  Options:\n    -e, --email TEXT     User email\n    -p, --password TEXT  User password\n    --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n    -t, --token TEXT     User token\n    -h, --help           Show this message and exit.\n  \n  Commands:\n    artists       List artists\n    bests         Generate bests playlists with some rules\n    clean         Clean all musics\n    consistency   Check music files consistency\n    execute       Raw query\n    filter        Print a filter\n    filters       List filters\n    find          Just list music files\n    folders       List folders\n    help          Print help\n    load-filters  Load default filters\n    play          Music player\n    playlist      Generate a new playlist\n    scan          (re)Load musics\n    stats         Generate some stats for music collection with filters\n    sync          Copy selected musics with filters to destination folder\n    watch         Watch files changes in folders\n\n\nmusicbot local artists\n**********************\n.. code-block::\n\n  Usage: musicbot local artists [OPTIONS]\n  \n    List artists\n  \n  Options:\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    -h, --help                     Show this message and exit.\n\n\nmusicbot local bests\n********************\n.. code-block::\n\n  Usage: musicbot local bests [OPTIONS] PATH\n  \n    Generate bests playlists with some rules\n  \n  Options:\n    --dry                   Take no real action  [default: False]\n    --name TEXT             Filter name\n    --limit INTEGER         Fetch a maximum limit of music\n    --youtubes TEXT         Select musics with a youtube link\n    --no-youtubes TEXT      Select musics without youtube link\n    --spotifys TEXT         Select musics with a spotifys link\n    --no-spotifys TEXT      Select musics without spotifys link\n    --formats TEXT          Select musics with file format\n    --no-formats TEXT       Filter musics without format\n    --keywords TEXT         Select musics with keywords\n    --no-keywords TEXT      Filter musics without keywords\n    --artists TEXT          Select musics with artists\n    --no-artists TEXT       Filter musics without artists\n    --albums TEXT           Select musics with albums\n    --no-albums TEXT        Filter musics without albums\n    --titles TEXT           Select musics with titles\n    --no-titles TEXT        Filter musics without titless\n    --genres TEXT           Select musics with genres\n    --no-genres TEXT        Filter musics without genres\n    --min-duration INTEGER  Minimum duration filter (hours:minutes:seconds)\n    --max-duration INTEGER  Maximum duration filter (hours:minutes:seconds))\n    --min-size INTEGER      Minimum file size filter (in bytes)\n    --max-size INTEGER      Maximum file size filter (in bytes)\n    --min-rating FLOAT      Minimum rating  [default: 0.0]\n    --max-rating FLOAT      Maximum rating  [default: 5.0]\n    --relative              Generate relatives paths\n    --shuffle               Randomize selection\n    --prefix TEXT           Append prefix before each path (implies relative)\n    --suffix TEXT           Append this suffix to playlist name\n    -h, --help              Show this message and exit.\n\n\nmusicbot local clean\n********************\n.. code-block::\n\n  Usage: musicbot local clean [OPTIONS]\n  \n    Clean all musics\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot local consistency\n**************************\n.. code-block::\n\n  Usage: musicbot local consistency [OPTIONS] [FOLDERS]...\n  \n    Check music files consistency\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot local execute\n**********************\n.. code-block::\n\n  Usage: musicbot local execute [OPTIONS] QUERY\n  \n    Raw query\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot local filter\n*********************\n.. code-block::\n\n  Usage: musicbot local filter [OPTIONS] NAME\n  \n    Print a filter\n  \n  Options:\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    -h, --help                     Show this message and exit.\n\n\nmusicbot local filters\n**********************\n.. code-block::\n\n  Usage: musicbot local filters [OPTIONS]\n  \n    List filters\n  \n  Options:\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    -h, --help                     Show this message and exit.\n\n\nmusicbot local find\n*******************\n.. code-block::\n\n  Usage: musicbot local find [OPTIONS] [FOLDERS]...\n  \n    Just list music files\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot local folders\n**********************\n.. code-block::\n\n  Usage: musicbot local folders [OPTIONS]\n  \n    List folders\n  \n  Options:\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    -h, --help                     Show this message and exit.\n\n\nmusicbot local load-filters\n***************************\n.. code-block::\n\n  Usage: musicbot local load-filters [OPTIONS]\n  \n    Load default filters\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot local play\n*******************\n.. code-block::\n\n  Usage: musicbot local play [OPTIONS]\n  \n    Music player\n  \n  Options:\n    -e, --email TEXT        User email\n    -p, --password TEXT     User password\n    --graphql TEXT          GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n    -t, --token TEXT        User token\n    --name TEXT             Filter name\n    --limit INTEGER         Fetch a maximum limit of music\n    --youtubes TEXT         Select musics with a youtube link\n    --no-youtubes TEXT      Select musics without youtube link\n    --spotifys TEXT         Select musics with a spotifys link\n    --no-spotifys TEXT      Select musics without spotifys link\n    --formats TEXT          Select musics with file format\n    --no-formats TEXT       Filter musics without format\n    --keywords TEXT         Select musics with keywords\n    --no-keywords TEXT      Filter musics without keywords\n    --artists TEXT          Select musics with artists\n    --no-artists TEXT       Filter musics without artists\n    --albums TEXT           Select musics with albums\n    --no-albums TEXT        Filter musics without albums\n    --titles TEXT           Select musics with titles\n    --no-titles TEXT        Filter musics without titless\n    --genres TEXT           Select musics with genres\n    --no-genres TEXT        Filter musics without genres\n    --min-duration INTEGER  Minimum duration filter (hours:minutes:seconds)\n    --max-duration INTEGER  Maximum duration filter (hours:minutes:seconds))\n    --min-size INTEGER      Minimum file size filter (in bytes)\n    --max-size INTEGER      Maximum file size filter (in bytes)\n    --min-rating FLOAT      Minimum rating  [default: 0.0]\n    --max-rating FLOAT      Maximum rating  [default: 5.0]\n    --relative              Generate relatives paths\n    --shuffle               Randomize selection\n    -h, --help              Show this message and exit.\n\n\nmusicbot local playlist\n***********************\n.. code-block::\n\n  Usage: musicbot local playlist [OPTIONS] [PATH]\n  \n    Generate a new playlist\n  \n  Options:\n    --dry                          Take no real action  [default: False]\n    --name TEXT                    Filter name\n    --limit INTEGER                Fetch a maximum limit of music\n    --youtubes TEXT                Select musics with a youtube link\n    --no-youtubes TEXT             Select musics without youtube link\n    --spotifys TEXT                Select musics with a spotifys link\n    --no-spotifys TEXT             Select musics without spotifys link\n    --formats TEXT                 Select musics with file format\n    --no-formats TEXT              Filter musics without format\n    --keywords TEXT                Select musics with keywords\n    --no-keywords TEXT             Filter musics without keywords\n    --artists TEXT                 Select musics with artists\n    --no-artists TEXT              Filter musics without artists\n    --albums TEXT                  Select musics with albums\n    --no-albums TEXT               Filter musics without albums\n    --titles TEXT                  Select musics with titles\n    --no-titles TEXT               Filter musics without titless\n    --genres TEXT                  Select musics with genres\n    --no-genres TEXT               Filter musics without genres\n    --min-duration INTEGER         Minimum duration filter (hours:minutes:seconds)\n    --max-duration INTEGER         Maximum duration filter (hours:minutes:seconds))\n    --min-size INTEGER             Minimum file size filter (in bytes)\n    --max-size INTEGER             Maximum file size filter (in bytes)\n    --min-rating FLOAT             Minimum rating  [default: 0.0]\n    --max-rating FLOAT             Maximum rating  [default: 5.0]\n    --relative                     Generate relatives paths\n    --shuffle                      Randomize selection\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    -h, --help                     Show this message and exit.\n\n\nmusicbot local scan\n*******************\n.. code-block::\n\n  Usage: musicbot local scan [OPTIONS] [FOLDERS]...\n  \n    (re)Load musics\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot local stats\n********************\n.. code-block::\n\n  Usage: musicbot local stats [OPTIONS]\n  \n    Generate some stats for music collection with filters\n  \n  Options:\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    --name TEXT                    Filter name\n    --limit INTEGER                Fetch a maximum limit of music\n    --youtubes TEXT                Select musics with a youtube link\n    --no-youtubes TEXT             Select musics without youtube link\n    --spotifys TEXT                Select musics with a spotifys link\n    --no-spotifys TEXT             Select musics without spotifys link\n    --formats TEXT                 Select musics with file format\n    --no-formats TEXT              Filter musics without format\n    --keywords TEXT                Select musics with keywords\n    --no-keywords TEXT             Filter musics without keywords\n    --artists TEXT                 Select musics with artists\n    --no-artists TEXT              Filter musics without artists\n    --albums TEXT                  Select musics with albums\n    --no-albums TEXT               Filter musics without albums\n    --titles TEXT                  Select musics with titles\n    --no-titles TEXT               Filter musics without titless\n    --genres TEXT                  Select musics with genres\n    --no-genres TEXT               Filter musics without genres\n    --min-duration INTEGER         Minimum duration filter (hours:minutes:seconds)\n    --max-duration INTEGER         Maximum duration filter (hours:minutes:seconds))\n    --min-size INTEGER             Minimum file size filter (in bytes)\n    --max-size INTEGER             Maximum file size filter (in bytes)\n    --min-rating FLOAT             Minimum rating  [default: 0.0]\n    --max-rating FLOAT             Maximum rating  [default: 5.0]\n    --relative                     Generate relatives paths\n    --shuffle                      Randomize selection\n    -h, --help                     Show this message and exit.\n\n\nmusicbot local sync\n*******************\n.. code-block::\n\n  Usage: musicbot local sync [OPTIONS] DESTINATION\n  \n    Copy selected musics with filters to destination folder\n  \n  Options:\n    --dry                   Take no real action  [default: False]\n    --name TEXT             Filter name\n    --limit INTEGER         Fetch a maximum limit of music\n    --youtubes TEXT         Select musics with a youtube link\n    --no-youtubes TEXT      Select musics without youtube link\n    --spotifys TEXT         Select musics with a spotifys link\n    --no-spotifys TEXT      Select musics without spotifys link\n    --formats TEXT          Select musics with file format\n    --no-formats TEXT       Filter musics without format\n    --keywords TEXT         Select musics with keywords\n    --no-keywords TEXT      Filter musics without keywords\n    --artists TEXT          Select musics with artists\n    --no-artists TEXT       Filter musics without artists\n    --albums TEXT           Select musics with albums\n    --no-albums TEXT        Filter musics without albums\n    --titles TEXT           Select musics with titles\n    --no-titles TEXT        Filter musics without titless\n    --genres TEXT           Select musics with genres\n    --no-genres TEXT        Filter musics without genres\n    --min-duration INTEGER  Minimum duration filter (hours:minutes:seconds)\n    --max-duration INTEGER  Maximum duration filter (hours:minutes:seconds))\n    --min-size INTEGER      Minimum file size filter (in bytes)\n    --max-size INTEGER      Maximum file size filter (in bytes)\n    --min-rating FLOAT      Minimum rating  [default: 0.0]\n    --max-rating FLOAT      Maximum rating  [default: 5.0]\n    --relative              Generate relatives paths\n    --shuffle               Randomize selection\n    -h, --help              Show this message and exit.\n\n\nmusicbot local watch\n********************\n.. code-block::\n\n  Usage: musicbot local watch [OPTIONS]\n  \n    Watch files changes in folders\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot music\n**************\n.. code-block::\n\n  Usage: musicbot music [OPTIONS] COMMAND [ARGS]...\n  \n    Music file\n  \n  Options:\n    -h, --help  Show this message and exit.\n  \n  Commands:\n    fingerprint  Print music fingerprint\n    help         Print help\n\n\nmusicbot music fingerprint\n**************************\n.. code-block::\n\n  Usage: musicbot music fingerprint [OPTIONS] PATH\n  \n    Print music fingerprint\n  \n  Options:\n    --acoustid-api-key TEXT  AcoustID API Key\n    -h, --help               Show this message and exit.\n\n\nmusicbot spotify\n****************\n.. code-block::\n\n  Usage: musicbot spotify [OPTIONS] COMMAND [ARGS]...\n  \n    Spotify tool\n  \n  Options:\n    -h, --help  Show this message and exit.\n  \n  Commands:\n    help       Print help\n    playlist   Show playlist\n    playlists  List playlists\n    tracks     Show tracks\n\n\nmusicbot spotify playlist\n*************************\n.. code-block::\n\n  Usage: musicbot spotify playlist [OPTIONS] NAME\n  \n    Show playlist\n  \n  Options:\n    --spotify TEXT  Spotify token\n    -h, --help      Show this message and exit.\n\n\nmusicbot spotify playlists\n**************************\n.. code-block::\n\n  Usage: musicbot spotify playlists [OPTIONS]\n  \n    List playlists\n  \n  Options:\n    --spotify TEXT  Spotify token\n    -h, --help      Show this message and exit.\n\n\nmusicbot spotify tracks\n***********************\n.. code-block::\n\n  Usage: musicbot spotify tracks [OPTIONS]\n  \n    Show tracks\n  \n  Options:\n    --spotify TEXT                 Spotify token\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    -h, --help                     Show this message and exit.\n\n\nmusicbot user\n*************\n.. code-block::\n\n  Usage: musicbot user [OPTIONS] COMMAND [ARGS]...\n  \n    User management\n  \n  Options:\n    -h, --help  Show this message and exit.\n  \n  Commands:\n    help                        Print help\n    list                        List users (admin)\n    login (token)               Authenticate user\n    register (add,create,new)   Register a new user\n    unregister (delete,remove)  Remove a user\n\n\nmusicbot user list\n******************\n.. code-block::\n\n  Usage: musicbot user list [OPTIONS]\n  \n    List users (admin)\n  \n  Options:\n    --output [table|json|m3u|csv]  Output format  [default: table]\n    --graphql-admin TEXT           GraphQL endpoint  [default: http://127.0.0.1:5001/graphql]\n    -h, --help                     Show this message and exit.\n\n\nmusicbot user login\n*******************\n.. code-block::\n\n  Usage: musicbot user login [OPTIONS]\n  \n    Authenticate user\n  \n  Options:\n    -e, --email TEXT     User email\n    -p, --password TEXT  User password\n    --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n    -s, --save           Save to config file  [default: False]\n    -h, --help           Show this message and exit.\n\n\nmusicbot user register\n**********************\n.. code-block::\n\n  Usage: musicbot user register [OPTIONS]\n  \n    Register a new user\n  \n  Options:\n    -e, --email TEXT     User email\n    -p, --password TEXT  User password\n    --first-name TEXT    User first name\n    --last-name TEXT     User last name\n    --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n    -s, --save           Save to config file  [default: False]\n    -h, --help           Show this message and exit.\n\n\nmusicbot user unregister\n************************\n.. code-block::\n\n  Usage: musicbot user unregister [OPTIONS]\n  \n    Remove a user\n  \n  Options:\n    -e, --email TEXT     User email\n    -p, --password TEXT  User password\n    --graphql TEXT       GraphQL endpoint  [default: http://127.0.0.1:5000/graphql]\n    -t, --token TEXT     User token\n    -h, --help           Show this message and exit.\n\n\nmusicbot version\n****************\n.. code-block::\n\n  Usage: musicbot version [OPTIONS]\n  \n    Print version\n  \n    Equivalent : -V\n  \n  Options:\n    -h, --help  Show this message and exit.\n\n\nmusicbot youtube\n****************\n.. code-block::\n\n  Usage: musicbot youtube [OPTIONS] COMMAND [ARGS]...\n  \n    Youtube tool\n  \n  Options:\n    -h, --help  Show this message and exit.\n  \n  Commands:\n    find         Search a youtube link with artist and title\n    fingerprint  Fingerprint a youtube video\n    help         Print help\n    search       Search a youtube link with artist and title\n\n\nmusicbot youtube find\n*********************\n.. code-block::\n\n  Usage: musicbot youtube find [OPTIONS] PATH\n  \n    Search a youtube link with artist and title\n  \n  Options:\n    --youtube-api-key TEXT   YouTube API Key\n    --acoustid-api-key TEXT  AcoustID API Key\n    -h, --help               Show this message and exit.\n\n\nmusicbot youtube fingerprint\n****************************\n.. code-block::\n\n  Usage: musicbot youtube fingerprint [OPTIONS] URL\n  \n    Fingerprint a youtube video\n  \n  Options:\n    --acoustid-api-key TEXT  AcoustID API Key\n    -h, --help               Show this message and exit.\n\n\nmusicbot youtube search\n***********************\n.. code-block::\n\n  Usage: musicbot youtube search [OPTIONS] ARTIST TITLE\n  \n    Search a youtube link with artist and title\n  \n  Options:\n    --youtube-api-key TEXT  YouTube API Key\n    -h, --help              Show this message and exit.\n\n\n',
    'author': 'Adrien Pensart',
    'author_email': 'crunchengine@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import isodate
import click
import requests
from humanfriendly import format_timespan
from musicbot.helpers import config_string

logger = logging.getLogger(__name__)
logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)
logging.getLogger('googleapiclient.discovery').setLevel(logging.CRITICAL)

YOUTUBE_API_SERVICE_NAME = "youtube"
YOUTUBE_API_VERSION = "v3"

DEFAULT_YOUTUBE_API_KEY = None
youtube_api_key_option = [click.option('--youtube-api-key', help='YouTube API Key', default=DEFAULT_YOUTUBE_API_KEY, callback=config_string)]


def youtube_duration(dur):
    if dur is None:
        return "any"
    if 0 <= dur <= 4 * 60:
        return "short"
    if 4 * 60 < dur <= 20 * 60:
        return "medium"
    if dur > 20 * 60:
        return "long"
    return "any"


# pylint: disable-msg=too-many-locals
def search(youtube_api_key, artist, title, duration=None):
    try:
        string = ' '.join([artist, title])
        parsingChannelUrl = "https://www.googleapis.com/youtube/v3/search"
        parsingChannelHeader = {'cache-control': "no-cache"}
        parsingChannelQueryString = {"part": "id,snippet", "maxResults": "10",
                                     "key": youtube_api_key, "type": "video", "q": string, "safeSearch": "none", "videoDuration": youtube_duration(duration)}
        parsingChannel = None

        resp = requests.get(parsingChannelUrl, headers=parsingChannelHeader, params=parsingChannelQueryString)
        parsingChannel = resp.json()

        parsingChannelItems = parsingChannel.get("items")
        if parsingChannelItems is None or not parsingChannelItems:
            return []
        VideoIds = ",".join(str(x.get("id").get("videoId")) for x in parsingChannelItems)

        parsingVideoUrl = "https://www.googleapis.com/youtube/v3/videos"
        parsingVideoHeader = {'cache-control': "no-cache"}
        parsingVideoQueryString = {"part": 'id,snippet,contentDetails', "id": VideoIds, "key": youtube_api_key}
        parsingVideo = None

        resp = requests.get(parsingVideoUrl, headers=parsingVideoHeader, params=parsingVideoQueryString)
        parsingVideo = resp.json()
        results = parsingVideo.get("items")
        if results is None:
            return []

        for r in results:
            logger.info('%s %s', r["snippet"]["title"], format_timespan(isodate.parse_duration(r["contentDetails"]["duration"]).total_seconds()))
        if duration:
            mapping = {"https://www.youtube.com/watch?v=" + r["id"]: abs(isodate.parse_duration(r["contentDetails"]["duration"]).total_seconds() - duration) for r in results}
            links = sorted(mapping, key=mapping.get)
            return links
        return ["https://www.youtube.com/watch?v=" + r["id"] for r in results]
    except Exception as e:  # pylint: disable=broad-except
        logger.info(type(e))
        logger.info('Cannot find video for artist: %s title: %s duration: %s', artist, title, duration)
        return []



#!/usr/bin/env python3
import sys
import os
import logging
import click
import requests
import spotipy
from click.formatting import HelpFormatter
from attrdict import AttrDict
from musicbot import lib, helpers, config
from musicbot.user import MusicbotError

# little hacky but prevent click from rewraping
HelpFormatter.write_dl.__defaults__ = (50, 2)

bin_folder = os.path.dirname(__file__)
commands_folder = 'commands'
plugin_folder = os.path.join(bin_folder, commands_folder)
CONTEXT_SETTINGS = {'max_content_width': 140, 'terminal_width': 140, 'auto_envvar_prefix': 'MB', 'help_option_names': ['-h', '--help']}
logger = logging.getLogger('musicbot')


# import __version__ string
_version = {}
with open(os.path.join(bin_folder, "version.py")) as fp:
    exec(fp.read(), _version)  # pylint: disable=exec-used
__version__ = _version['__version__']
prog_name = "musicbot"


class SubCommandLineInterface(helpers.GroupWithHelp):
    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(plugin_folder):
            if filename.endswith('.py') and '__init__' not in filename:
                rv.append(filename[:-3])
        all_commands = rv + super().list_commands(ctx)
        all_commands.sort()
        return all_commands

    def get_command(self, ctx, cmd_name):
        ns = {}
        fn = os.path.join(plugin_folder, cmd_name + '.py')
        try:
            with open(fn) as f:
                code = compile(f.read(), fn, 'exec')
                ns['__name__'] = f'{commands_folder}.{cmd_name}'
                ns['__file__'] = fn
                eval(code, ns, ns)  # pylint: disable=eval-used
        except FileNotFoundError:
            return super().get_command(ctx, cmd_name)
        return ns['cli']


@click.group(cls=SubCommandLineInterface, context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.version_option(__version__, "--version", "-V", prog_name=prog_name)
@helpers.add_options(config.options)
@click.pass_context
def cli(ctx, **kwargs):
    """Music swiss knife, new gen."""
    ctx.obj = AttrDict
    ctx.obj.folder = bin_folder
    config.config.set(**kwargs)
    ctx.obj.config = config.config


@cli.command(short_help='Print version')
def version():
    '''Print version

       Equivalent : -V
    '''
    print(f"{prog_name}, version {__version__}")


def main(**kwargs):
    try:
        return cli.main(prog_name=prog_name, **kwargs)
    except (MusicbotError, spotipy.client.SpotifyException, requests.exceptions.ConnectionError) as e:
        logger.error(e)
        sys.exit(1)


if __name__ == '__main__':
    lib.raise_limits()
    main()
